SuperStrict

Rem
bbdoc: Holzchopf/MaxGuiLayout
about: Simple layout manager for MaxGUI
End Rem
Module Holzchopf.MaxGuiLayout

ModuleInfo "Version: 1.01"
ModuleInfo "Author: Holzchopf"
ModuleInfo "License: Public Domain"

Import MaxGUI.Drivers

Rem
bbdoc: Layout manager type
about: Every part of a layout is considered to be a layout manager itself
End Rem
Type TMaxGuiLayout
	Rem
	bbdoc: Parent layout manager
	End Rem
	Field parent:TMaxGuiLayout
	
	Rem
	bbdoc: Gadget of layout
	End Rem
	Field gadget:TGadget
	
	Field x:Int
	Field y:Int
	Field w:Int
	Field h:Int
	
	Field hspace:Int
	Field vspace:Int
	
	Field hmargin:Int
	Field vmargin:Int

	Field labelcorr:Int = 2

	Rem
	bbdoc: Two dimensional array of children
	End Rem
	Field sub:TMaxGuiLayout[,]
	
	Rem
	bbdoc: Creates a new layout
	about: Creates a new top-level layout manager which can contain several sub layouts.<br />
Parameters are as follows:
<table>
<tr>
	<th>Parameter</th><th>Description</th><th>Default</th>
</tr><tr>
	<td>pX</td><td>X position</td><td></td>
</tr><tr>
	<td>pY</td><td>Y position</td><td></td>
</tr><tr>
	<td>pW</td><td>Width</td><td></td>
</tr><tr>
	<td>pH</td><td>Height</td><td></td>
</tr><tr>
	<td>pHMargin</td><td>Horizontal margin</td><td>5</td>
</tr><tr>
	<td>pVMargin</td><td>Vertical margin</td><td>5</td>
</tr><tr>
	<td>pHSpace</td><td>Horizontal space between children</td><td>5</td>
</tr><tr>
	<td>pVSpace</td><td>Vertical space between children</td><td>5</td>
</tr><tr>
	<td>pParent</td><td>Parent #TMaxGuiLayout</td><td>Null</td>
</tr>
</table>
	End Rem
	Function Create:TMaxGuiLayout(pX:Int, pY:Int, pW:Int, pH:Int, ..
	 pHMargin:Int=5, pVMargin:Int=5, ..
	 pHSpace:Int=5, pVSpace:Int=5, ..
	 pParent:TMaxGuiLayout=Null)
		Local layout:TMaxGuiLayout = New TMaxGuiLayout
		layout.x = pX
		layout.y = pY
		layout.w = pW
		layout.h = pH
		layout.hspace = pHSpace
		layout.vspace = pVSpace
		layout.hmargin = pHMargin
		layout.vmargin = pVMargin
		layout.parent = pParent
		Return layout
	End Function
	
	Rem
	bbdoc: Splits a layout into pieces of given widths and heights
	about: pWidths and pHeights are Int arrays which define the size of each row and column.
The layout will be splitted horizontally and vertically in as much elements as are defined in the arrays respectively.
A size definition of 0 will cause the row or column to be sized automatically.<br />
If pGadget is true, a bounding panel gadget will automatically be created
(sometimes necessary in scalable windows).
	End Rem
	Method Split(pWidths:Int[], pHeights:Int[], pGadget:Int=False)
		sub = New TMaxGuiLayout[pWidths.length, pHeights.length]
		If gadget=Null
			If pGadget Then Gadget = Panel()
		Else
			w = gadget.ClientWidth()
			h = gadget.ClientHeight()
		EndIf
		_AutoSize(pWidths, w -2*hmargin -(pWidths.length-1)*hspace)
		_AutoSize(pHeights, h -2*vmargin -(pHeights.length-1)*vspace)
		Local _y:Int = vmargin
		If gadget = Null _y :+ y
		For Local v:Int = 0 Until pHeights.length
			Local _h:Int = pHeights[v]
			Local _x:Int = hmargin
			If gadget = Null _x :+ x
			For Local u:Int = 0 Until pWidths.length
				Local _w:Int = pWidths[u]
				sub[u,v] = Create(_x,_y,_w,_h, 0,0, hspace,vspace, Self)
				_x :+ _w +hspace
			Next
			_y :+ _h +vspace
		Next
	End Method
	
	Rem
	bbdoc: Creates a sub layout (usefull if nesting gadgets)
	End Rem
	Method Inner:TMaxGuiLayout(pWidth:Int=0, pHeight:Int=0)
		Split([pWidth],[pHeight])
		Return sub[0,0]
	End Method
	
	Rem
	bbdoc: Adds sub layouts on the right
	End Rem
	Method AddRight(pWidth:Int, pCount:Int=1)
		pCount = Max(1, pCount)
		If Not sub
			Local _widths:Int[] = [pWidth]
			For Local a:Int = 1 Until pCount
				_widths :+ [pWidth]
			Next
			Split(_widths,[0])
			Return
		EndIf
		Local sx:Int = sub.Dimensions()[0]
		Local sy:Int = sub.Dimensions()[1]
		Local nsub:TMaxGuiLayout[,] = New TMaxGuiLayout[sx+pCount,sy]
		For Local v:Int = 0 Until sy
			For Local u:Int = 0 Until sx
				nsub[u,v] = sub[u,v]
			Next
			Local _x:Int = sub[sx-1,v].x +sub[sx-1,v].w +hspace
			Local _y:Int = sub[sx-1,v].y
			Local _h:Int = sub[sx-1,v].h
			For Local a:Int = 0 Until pCount
				nsub[sx+a,v] = Create(_x, _y, pWidth, _h, 0,0, hspace,vspace, Self)
				_x :+ (pWidth +hspace)
			Next
		Next
		sub = nsub
	End Method

	Rem
	bbdoc: Adds sub layouts at the bottom
	End Rem
	Method AddBottom(pHeight:Int, pCount:Int=1)
		pCount = Max(1, pCount)
		If Not sub
			Local _heights:Int[] = [pHeight]
			For Local a:Int = 1 Until pCount
				_heights :+ [pHeight]
			Next
			Split([0],_heights)
			Return
		EndIf
		Local sx:Int = sub.Dimensions()[0]
		Local sy:Int = sub.Dimensions()[1]
		Local nsub:TMaxGuiLayout[,] = New TMaxGuiLayout[sx,sy+pCount]
		For Local u:Int = 0 Until sx
			For Local v:Int = 0 Until sy
				nsub[u,v] = sub[u,v]
			Next
			Local _x:Int = sub[u,sy-1].x
			Local _y:Int = sub[u,sy-1].y +sub[u,sy-1].h +vspace
			Local _w:Int = sub[u,sy-1].w
			For Local a:Int = 0 Until pCount
				nsub[u,sy+a] = Create(_x, _y, _w, pHeight, 0,0, hspace,vspace, Self)
				_y :+ (pHeight +vspace)
			Next
		Next
		sub = nsub
	End Method

	Rem
	bbdoc: Removes sub layouts from the right side
	End Rem
	Method RemoveRight(pCount:Int=1)
		pCount = Max(1, pCount)
		If Not sub Return
		Local sx:Int = sub.Dimensions()[0]
		Local sy:Int = sub.Dimensions()[1]
		Local ex:Int = sx -pCount
		Local nsub:TMaxGuiLayout[,] = New TMaxGuiLayout[sx-pCount,sy]
		For Local v:Int = 0 Until sy
			For Local u:Int = 0 Until sx
				If u<ex
					nsub[u,v] = sub[u,v]
				Else
					sub[u,v]._Free()
				EndIf
			Next
		Next
		sub = nsub
	End Method

	Rem
	bbdoc: Removes sub layouts from the bottom
	End Rem
	Method RemoveBottom(pCount:Int=1)
		pCount = Max(1, pCount)
		If Not sub Return
		Local sx:Int = sub.Dimensions()[0]
		Local sy:Int = sub.Dimensions()[1]
		Local ey:Int = sy -pCount
		Local nsub:TMaxGuiLayout[,] = New TMaxGuiLayout[sx,sy-pCount]
		For Local u:Int = 0 Until sx
			For Local v:Int = 0 Until sy
				If v<ey
					nsub[u,v] = sub[u,v]
				Else
					sub[u,v]._Free()
				EndIf
			Next
		Next
		sub = nsub
	End Method

	Method _Free()
		For Local _sub:TMaxGuiLayout = EachIn sub
			_sub._Free()
		Next
		If gadget
			FreeGadget(gadget)
		EndIf
		sub = Null
	End Method
	
	Method _AutoSize(pSizes:Int[] Var, pTotSize:Int)
		' total size
		Local stot:Int = 0
		Local sauto:Int = 0
		For Local _s:Int = EachIn pSizes
			stot :+ _s
			If _s=0 sauto :+1
		Next
		If sauto=0 Return
		' auto size
		For Local a:Int = 0 Until pSizes.length
			If pSizes[a]=0
				pSizes[a] = (pTotSize -stot) /sauto
				sauto :-1
				If sauto=0
					Return
				Else
					Exit
				EndIf
			EndIf
		Next
		' auto size rest
		_AutoSize(pSizes, pTotSize)
	End Method

	Rem
	bbdoc: Resizes a layout to fit it's childs size
	End Rem
	Method FitToContents()
		Local dim:Int[] = sub.Dimensions()
		Local _w:Int = 2*hmargin +hspace*(dim[0]-1)
		Local _h:Int = 2*vmargin +vspace*(dim[1]-1)
		For Local u:Int = 0 Until dim[0]
			_w :+ sub[u,0].w
		Next
		For Local v:Int = 0 Until dim[1]
			_h :+ sub[0,v].h
		Next
		If gadget
			Local woffset:Int = gadget.width -gadget.ClientWidth()
			Local hoffset:Int = gadget.height -gadget.ClientHeight()
			w = _w +woffset
			h = _h +hoffset
			gadget.SetShape(gadget.xpos, gadget.ypos, w, h)
		Else
			w = _w
			h = _h
		EndIf
	End Method

	Rem
	bbdoc: Lock the layout
	about: pHAuto and pVAuto define, which column and row shall be scaled when resizing the window.
	End Rem
	Method LockLayout(pHAuto:Int=-1, pVAuto:Int=-1)
		If Not sub Return
		Local dim:Int[] = sub.Dimensions()
		If pHAuto = -1
			pHAuto = dim[0]-1
		EndIf
		If pVAuto = -1
			pVAuto = dim[1]-1
		EndIf
		For Local v:Int = 0 Until dim[1]
			For Local u:Int = 0 Until dim[0]
				If sub[u,v].gadget
					Local lf:Int = EDGE_ALIGNED
					Local rt:Int = 0
					Local tp:Int = EDGE_ALIGNED
					Local bt:Int = 0
					If u = pHAuto
						rt = EDGE_ALIGNED
					ElseIf u > pHAuto
						lf = 0
						rt = EDGE_ALIGNED
					EndIf
					If v = pVAuto
						bt = EDGE_ALIGNED
					ElseIf v > pVAuto
						tp = 0
						bt = EDGE_ALIGNED
					EndIf
					sub[u,v].gadget.SetLayout(lf, rt, tp, bt)
				EndIf
				sub[u,v].LockLayout()
			Next
		Next
	End Method
	
	Rem
	bbdoc: Set the vertical offset of label gadgets
	about: Label gadget usually need a vertical offset to mach nearby gadgets
	End Rem
	Method SetLabelOffset(pOffset:Int)
		labelcorr = pOffset
	End Method
	Rem
	bbdoc: Set margins of a layout
	End Rem
	Method SetMargin(pHMargin:Int, pVMargin:Int)
		hmargin = pHMargin
		vmargin = pVMargin
	End Method
	Rem
	bbdoc: Set spacings of a layout
	End Rem
	Method SetSpacing(pHSpace:Int, pVSpace:Int)
		hspace = pHSpace
		vspace = pVSpace
	End Method
	
	Rem
	bbdoc: Set specific gadget
	End Rem
	Method SetGadget:TGadget(pGadget:Object)
		gadget = TGadget(pGadget)
		If Not gadget Return Null
		gadget.SetShape(x,y, w,h)
		Return gadget
	End Method
	
	Rem
	bbdoc: Creates a window for a layout
	about: Most commonly used for top-level layouts
	End Rem
	Method Window:TGadget(pName:String, pX:Int=0, pY:Int=0, pGroup:TGadget=Null, pStyle:Int=0)
		gadget = CreateWindow(pName, pX, pY, w, h, pGroup, WINDOW_CLIENTCOORDS | pStyle)
		Return gadget
	End Method
	
	Rem
	bbdoc: Creates a gadget for a layout with the specified create function
	about: The create function must be of one of following types
<table>
<tr>
	<th>Parameter</th><th>Type</th><th>Example</th>
</tr><tr>
	<td>pFunc</td><td>TGadget(Int, Int, Int, Int, TGadget, Int)</td><td>CreateCanvas</td>
</tr><tr>
	<td>pFunc1</td><td>TGadget(String, Int, Int, Int, Int, TGadget, Int)</td><td>CreateButton</td>
</tr><tr>
	<td>pFunc2</td><td>TGadget(Int, Int, Int, Int, TGadget, Int, String)</td><td>CreatePanel</td>
</tr>
</table>
There are methods with increased usability like #Button, #Canvas and so on.
#_Gadget should only be used to create gadgets which have no TMaxGuiLayout-built-in create method, e.g: CreateScrollPanel
	End Rem
	Method _Gadget:TGadget(..
	  pFunc:TGadget(x:Int,y:Int,w:Int,h:Int,g:TGadget,s:Int)=Null, ..
	  pFunc1:TGadget(pLabel:String,x:Int,y:Int,w:Int,h:Int,g:TGadget,s:Int)=Null, ..
	  pFunc2:TGadget(x:Int,y:Int,w:Int,h:Int,g:TGadget,s:Int,pLabel:String)=Null, ..
	  pGroup:TGadget=Null, pStyle:Int=0, pLabel:String="")
		If pGroup=Null pGroup=GetParent()
		If pFunc
			gadget = pFunc(x,y,w,h, pGroup, pStyle)
		ElseIf pFunc1
			gadget = pFunc1(pLabel, x,y,w,h, pGroup, pStyle)
		Else
			gadget = pFunc2(x,y,w,h, pGroup, pStyle, pLabel)
		EndIf
		gadget.SetLayout(EDGE_ALIGNED, 0, EDGE_ALIGNED, 0)
		Return gadget
	End Method
	
	Rem
	bbdoc: Creates a button for that layout
	End Rem
	Method Button:TGadget(pLabel:String="", pGroup:TGadget=Null, pStyle:Int=0)
		Return _Gadget(,CreateButton,, pGroup, pStyle, pLabel)
	End Method
	Rem
	bbdoc: Creates a canvas for that layout
	End Rem
	Method Canvas:TGadget(pGroup:TGadget=Null, pStyle:Int=0)
		Return _Gadget(CreateCanvas,,, pGroup, pStyle)
	End Method
	Rem
	bbdoc: Creates a combo box for that layout
	End Rem
	Method ComboBox:TGadget(pGroup:TGadget=Null, pStyle:Int=0)
		Return _Gadget(CreateComboBox,,, pGroup, pStyle)
	End Method
	Rem
	bbdoc: Creates a HTML view for that layout
	End Rem
	Method HTMLView:TGadget(pGroup:TGadget=Null, pStyle:Int=0)
		Return _Gadget(CreateHTMLView,,, pGroup, pStyle)
	End Method
	Rem
	bbdoc: Creates a label for that layout
	End Rem
	Method Label:TGadget(pLabel:String, pGroup:TGadget=Null, pStyle:Int=0)
		y :+ labelcorr
		Local g:TGadget = _Gadget(,CreateLabel,, pGroup, pStyle, pLabel)
		y :- labelcorr
		Return g
	End Method
	Rem
	bbdoc: Creates a list box for that layout
	End Rem
	Method ListBox:TGadget(pGroup:TGadget=Null, pStyle:Int=0)
		Return _Gadget(CreateListBox,,, pGroup, pStyle)
	End Method
	Rem
	bbdoc: Creates a panel for that layout
	End Rem
	Method Panel:TGadget(pLabel:String="", pGroup:TGadget=Null, pStyle:Int=0)
		Return _Gadget(,,CreatePanel, pGroup, pStyle, pLabel)
	End Method
	Rem
	bbdoc: Creates a prorgress bar for that layout
	End Rem
	Method ProgBar:TGadget(pGroup:TGadget=Null, pStyle:Int=0)
		Return _Gadget(CreateProgBar,,, pGroup, pStyle)
	End Method
	Rem
	bbdoc: Creates a slider for that layout
	End Rem
	Method Slider:TGadget(pGroup:TGadget=Null, pStyle:Int=0)
		Return _Gadget(CreateSlider,,, pGroup, pStyle)
	End Method
	Rem
	bbdoc: Creates a tabber for that layout
	End Rem
	Method Tabber:TGadget(pGroup:TGadget=Null, pStyle:Int=0)
		Return _Gadget(CreateTabber,,, pGroup, pStyle)
	End Method
	Rem
	bbdoc: Creates a text area for that layout
	End Rem
	Method TextArea:TGadget(pGroup:TGadget=Null, pStyle:Int=0)
		Return _Gadget(CreateTextArea,,, pGroup, pStyle)
	End Method
	Rem
	bbdoc: Creates a text field for that layout
	End Rem
	Method TextField:TGadget(pGroup:TGadget=Null, pStyle:Int=0)
		Return _Gadget(CreateTextField,,, pGroup, pStyle)
	End Method
	Rem
	bbdoc: Creates a tree view for that layout
	End Rem
	Method TreeView:TGadget(pGroup:TGadget=Null, pStyle:Int=0)
		Return _Gadget(CreateTreeView,,, pGroup, pStyle)
	End Method
	Rem
	bbdoc: Returns the parent gadget of a layout
	End Rem
	Method GetParent:TGadget()
		If Not parent Return Null
		If parent.gadget<>Null Return parent.gadget
		Return parent.GetParent()
	End Method
End Type

